<!-- Copyleft 2004 - see COPYING for details [LGPL] -->

<ibex>

    // Returns number of chars the pixel position pos is into text t.
    //
    // Always rounds down.
    //
    //  t     : string -- basis for character counting
    //  pxpos : int    -- pixel position on the text from string t
    //  pxlen : int    -- pixel width of text t (often box width)
    //  tfont : font   -- name of font used for width of text
    //  tsize : int    -- size of font used for width of text
    static.getpos = function(t, pxpos, pxlen, tfont, tsize) {
        // short circuit extremes
        if (0 >= pxpos) return 0;
        if (pxpos >= pxlen) return t.length;

        // inital guess based on average character width
        var guessch = ibex.math.min(t.length, ibex.math.floor(pxpos / (pxlen / t.length)));
        var guesspx = ibex.ui.font.width(tfont, tsize, t.substring(0, guessch));

        if (guesspx > pxpos) {
            while (guesspx > pxpos) {
                // textwidth of individual character must account for font kerning
                guesspx -= ibex.ui.font.width(tfont, tsize, t.substring(guessch -1, guessch +1));
                guesspx += ibex.ui.font.width(tfont, tsize, t.charAt(guessch));
                guessch--;
            }
        } else if (pxpos > guesspx) {
            while (pxpos > guesspx) {
                guessch++;
                if (guessch >= t.length) break;
                guesspx += ibex.ui.font.width(tfont, tsize, t.substring(guessch -1, guessch+1));
                guesspx -= ibex.ui.font.width(tfont, tsize, t.charAt(guessch));
            }
            guessch--;  // round down
        }

        return guessch;
    }


    <ui:box shrink="true">

        // add and remove cursor traps as the cursor box changes
        thisbox.cur ++= function(newcur) {
            if (cur) {
                cur.pos  --= cur_pos;
                cur.posx --= cur_posx_read;
                cur.posx --= cur_posx_write;
            }
            if (newcur) {
                newcur.pos  ++= cur_pos;
                newcur.posx ++= cur_posx_read;
                newcur.posx ++= cur_posx_write;
            }
        }

        // add and remove selection traps as the selection box changes
        thisbox.sel ++= function(newsel) {
            if (sel) {
                sel.pos1 --= sel_pos1;
                sel.pos2 --= sel_pos2;
            }
            if (newsel) {
                newsel.pos1 ++= sel_pos1;
                newsel.pos2 ++= sel_pos2;
            }
        }

        // designed to trap cur.pos -- updates position of the cursor
        // based on the character position stored in cur.pos
        var cur_pos = function(newpos) {
            //ibex.log.info("cur.pos trap put "+newpos+", text.length="+text.length);
            if (newpos > text.length) newpos = text.length;
            else if (0 > newpos) newpos = 0;

            if (newpos > 0) ibex.thread = function() {
                ibex.ui.font.wait(font, fontsize, text);
                cur.x = ibex.ui.font.width(font, fontsize, text.substring(0, newpos));
            } else cur.x = 0;

            if (focused) cur.visible = true;
        }

        // designed to trap cur.posx --
        // returns the pixel based position of the cursor
        var cur_posx_read = function() { return cur.x; }

        // designed to trap cur.posx -- updates cur.pos
        // based on the pixel value stored in cur.posx
        //
        // this two-step put to the x value is designed to normalise
        // the position of the cursor to being between characters
        var cur_posx_write = function(newx) {
            // calculate the cursor position in terms of characters
            ibex.thread = function() {
                ibex.ui.font.wait(font, fontsize, text);

                // short circuit limits
                if (0 >= newx) { cur.pos = 0; return; }
                if (newx > width) { cur.pos = text.length; return; }

                var pos = static.getpos(text, newx, width, font, fontsize);

                // getpos always rounds down see if we can get closer
                // to the mark by skipping over a char
                if (text.length > pos) {
                    var diff = newx - ibex.ui.font.width(font, fontsize, text.substring(0, pos));
                    if (diff > ibex.ui.font.width(font, fontsize, text.charAt(pos)) / 2) pos++;
                }
    
                cur.pos = pos;
            }
        }

        // designed to trap sel.pos1 -- resets the selection and
        // prepares it to start again from the given character position
        var sel_pos1 = function(v) {
            sel.text = null;
        }

        // designed to trap sel.pos2 -- obtains the second selection
        // position and 
        var sel_pos2 = function(v) {
            ibex.log.info("pos2 trapped, new value="+v+", (pos1="+sel.pos1+")");
            var p1, p2;
            if (v > sel.pos1) { p1 = sel.pos1; p2 = v; }
            else              { p2 = sel.pos1; p1 = v; }

            // update selected text
            sel.text = text.substring(p1, p2);
            sel.x = ibex.ui.font.width(font, fontsize, text.substring(0, p1));
        }

        // designed to be applied to thisbox.Move
        var move = function(v) {
            cur.posx = mouse.x;
            ibex.thread = function() { sel.pos2 = cur.pos; }
        }

        focused ++= function(v) {
            cur.visible = v;
        }

        Press1 ++= function(v) {
            ibex.log.info("Press1 in ibex.lib.text");
            cur.posx = mouse.x;
            // TODO: enable selection
            ibex.thread = function() {
                sel.pos1 = cur.pos;
                Move ++= move;
            }
        }

        // TODO: use surface.Release1
        Release1 ++= function(v) {
            ibex.log.info("Release1 in ibex.lib.text");
            Move --= move;
        }

        KeyPressed ++= function(v) {
            ibex.log.info("KeyPressed called with value: "+v);

            // TODO: check for casing and for prefix modifers (eg. C-Y, a-g)
            if (v.length > 1) switch (v) {
                // TODO
                case "left":  cur.pos--; return;
                case "C-LEFT":
                case "c-left": return;

                case "right": cur.pos++; return;
                case "C-RIGHT":
                case "c-right": return;

                case "home":  cur.pos = 0; return;
                case "end":   cur.pos = text.length; return;
                case "DELETE":
                case "back_space":
                    text = text.substring(0, cur.pos - 1) + text.substring(cur.pos);
                    cur.pos--;
                    return;
                case "BACK_SPACE":
                case "delete":
                    text = text.substring(0, cur.pos) + text.substring(cur.pos + 1);
                    return;
                case "shift":  ibex.log.warn("not-yet-implemented"); return;
                case "alt":    ibex.log.warn("not-yet-implemented"); return;
                case "ctrl":   ibex.log.warn("not-yet-implemented"); return;
                case "insert": ibex.log.warn("not-yet-implemented"); return;
                case "tab":    ibex.log.warn("not-yet-implemented"); return;
            }

            if (v.length == 1) {
                ibex.log.info("KeyPressed resulting in text insertion of "+v);
                text = text.substring(0, cur.pos) + v + text.substring(cur.pos);
                cur.pos += v.length;
            }
        }
   </ui:box>
</ibex>
