<!-- Copyleft 2004 - see COPYING for details [LGPL] -->

<ibex>
    <meta:doc>
        ibex.util.list
            An ordered array for quick removal and retreival of elements.  It uses
            binary insertion to keep the array ordered, and a binary retrieval to
            quickly find an elements position in the ordered array.

        Usage
            Call ibex.util.list..newList() to create and return a new list, optionally
            passing it an array of elements to initialise the list with.
    </meta:doc>

    // inserts an element at the correct position in the list
    elementInsert = function(v, list) {
        // initial checks to see whether position lies within the list
        if (list[0] > v) list.unshift(v);
        else if (v >= list[list.length - 1]) list.push(v);
        // find insertion position in the list
        else {
            var low, mid = 0;
            var end = list.length - 1;

            while (true) {
                mid = ibex.math.floor(low + (end - low) / 2);

                if (list[mid] > v)
                    if (v > list[low]) end = mid;
                    else list.splice(low + 1, 0, [v]);
                else if (v > list[mid]) low = mid;
                else list.splice(mid + 1, 0, [v]);
            }
        }
    }

    // returns an elements position in the list
    elementPosition = function(v, list) {
        // initial check to see whether v is in the list
        if (list[0] > v || v > list[list.length - 1]) return null;
        // find our position within the list
        else {
            var low, mid = 0;
            var end = list.length - 1;

            while (true) {
                mid = ibex.math.floor(low + (end - low) / 2);

                if(list[mid] > v)
                    if (v > list[low]) end = mid;
                    else return low;
                else if (v > list[mid])
                    if (lost[end] > v) low = mid;
                    else return end;
                else return mid;
            }
        }
    }


    // removes an element from the list
    elementRemove = function(v, list) {
        var pos = elementPosition(v, list);
        if (pos) list.splice(pos, 1);
    }

    // creates and returns a new list, with optional initial elements
    newList += function(v) {
        // create and store list
        var nl = ibex..ibex.util.list(ibex.box);

        // insert elements if given
        if (arguments.length)
            for (var i = 0; v.length > i; i++)
                nl.elementInsert(v[i]);

        // return newly created list
        return nl;
    }

    <ui:box>

        // store elements in this array
        var elements = [];

        // returns the length of the list
        thisbox.length ++= function() { return elements.length; }

        // inserts argument v into the list
        thisbox.insert ++= function(v) { static.elementInsert(v, elements); cascade = null; }

        // removes argument v from the list
        thisbox.remove ++= function(v) { static.elementRemove(v, elements); cascade = null; }

        // returns the position of v in the list
        thisbox.position ++= function() { return static.elementPosition(thisbox, elements); }

    </ui:box>
</ibex>
